package com.jeffhanson.resources;

import javax.servlet.ServletContext;
import javax.servlet.http.HttpServletRequest;
import java.net.URL;
import java.net.MalformedURLException;
import java.net.URLConnection;
import java.io.InputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;

/**
 * User: jhanson
 * Date: Jun 11, 2008
 * Time: 4:29:51 PM
 */
public class ResourceCache
{
  private static final String RESOURCE_FOLDER = "/WEB-INF/classes";

  /* todo - add logic to configure instance from dynamic data */
  private static ResourceCache instance = new ResourceCache();

  // =====================================================
  // non-public static methods
  // =====================================================

  private static ResourceCache getInstance()
  {
    return instance;
  }

  // =====================================================
  // public static methods
  // =====================================================

  public static void create(String resourcePath, ServletContext servletContext, HttpServletRequest request)
    throws ResourceException
  {
    getInstance().createResource(resourcePath, servletContext, request);
  }

  public static Resource read(String resourcePath, ServletContext servletContext)
    throws ResourceException
  {
    return getInstance().readResource(resourcePath, servletContext);
  }

  public static void update(String resourcePath, ServletContext servletContext, HttpServletRequest request)
    throws ResourceException
  {
    getInstance().updateResource(resourcePath, servletContext, request);
  }

  public static void delete(String resourcePath, ServletContext servletContext)
    throws ResourceException
  {
    getInstance().deleteResource(resourcePath, servletContext);
  }

  public static long getLastModified(String resourcePath, ServletContext servletContext)
    throws ResourceException
  {
    return getInstance().getResourceLastModified(resourcePath, servletContext);
  }

  // =====================================================
  // non-public instance methods
  // =====================================================

  private void createResource(String resourcePath, ServletContext servletContext, HttpServletRequest request)
    throws ResourceException
  {
    throw new ResourceException(new UnsupportedOperationException(getClass().getName()
                                                                        + ".createResource('" + resourcePath + "')"));
  }

  private Resource readResource(String resourcePath, ServletContext servletContext)
    throws ResourceException
  {
    if (resourcePath == null || resourcePath.length() == 0)
    {
      throw new ResourceException("Resource path is empty in "
                                        + getClass().getName()
                                        + ".readResource()");
    }

    // check that the resource exists
    //
    URL resourceURL = null;
    try
    {
      resourceURL = servletContext.getResource(RESOURCE_FOLDER + resourcePath);
    }
    catch (MalformedURLException e)
    {
      throw new ResourceException(e);
    }

    if (resourceURL == null)
    {
      throw new ResourceException("Resource [" + resourcePath + "] not found in "
                                        + getClass().getName()
                                        + ".readResource()");
    }

    // get the file name of the resource
    //
    final String resourceFileName = resourceURL.getFile();

    // get the contents of the resource file
    //
    InputStream inStream = servletContext.getResourceAsStream(RESOURCE_FOLDER + resourcePath);
    ByteArrayOutputStream outStream = new ByteArrayOutputStream();
    try
    {
      byte[] bytes = new byte[4096];
      int bytesRead = 0;
      while ((bytesRead = inStream.read(bytes)) > 0)
      {
        outStream.write(bytes, 0, bytesRead);
      }
    }
    catch (IOException e)
    {
      throw new ResourceException("Error retrieving resource [" + resourcePath + "] in "
                                        + getClass().getName()
                                        + ".readResource()");
    }

    final byte[] resourceBytes = outStream.toByteArray();

    // create the resource to return
    //
    return new Resource()
    {
      public String getMimeType()
      {
        if (resourceFileName.toLowerCase().endsWith(".txt"))
        {
          return("text/plain");
        }
        else if (resourceFileName.toLowerCase().endsWith(".xml"))
        {
          return("text/xml");
        }
        else if (resourceFileName.toLowerCase().endsWith(".html"))
        {
          return("text/html");
        }
        else if (resourceFileName.toLowerCase().endsWith(".rss"))
        {
          return("application/rss+xml");
        }
        else if (resourceFileName.toLowerCase().endsWith(".rdf"))
        {
          return("application/rdf+xml");
        }
        else if (resourceFileName.toLowerCase().endsWith(".atom"))
        {
          return("application/atom+xml");
        }
        else if (resourceFileName.toLowerCase().endsWith(".xls"))
        {
          return("application/vnd.ms-excel");
        }
        else if (resourceFileName.toLowerCase().endsWith(".doc"))
        {
          return("application/msword");
        }

        return "text/plain";
      }

      public int getContentLength()
      {
        System.out.println(getClass().getName() + ".getContentLength returning: " + resourceBytes.length);
        return resourceBytes.length;
      }

      public byte[] getBytes()
      {
        return resourceBytes;
      }

      public boolean canBeStringified()
      {
        if (resourceFileName.toLowerCase().endsWith(".txt"))
        {
          return true;
        }
        else if (resourceFileName.toLowerCase().endsWith(".xml"))
        {
          return true;
        }
        else if (resourceFileName.toLowerCase().endsWith(".html"))
        {
          return true;
        }
        else if (resourceFileName.toLowerCase().endsWith(".rss"))
        {
          return true;
        }
        else if (resourceFileName.toLowerCase().endsWith(".rdf"))
        {
          return true;
        }
        else if (resourceFileName.toLowerCase().endsWith(".atom"))
        {
          return true;
        }

        return false;
      }

      public String toString()
      {
        return new String(resourceBytes);
      }
    };
  }

  private void updateResource(String resourcePath, ServletContext servletContext, HttpServletRequest request)
    throws ResourceException
  {
    throw new ResourceException(new UnsupportedOperationException(getClass().getName()
                                                                        + ".updateResource('" + resourcePath + "')"));
  }

  private void deleteResource(String resourcePath, ServletContext servletContext)
    throws ResourceException
  {
    throw new ResourceException(new UnsupportedOperationException(getClass().getName()
                                                                        + ".deleteResource('" + resourcePath + "')"));
  }

  private long getResourceLastModified(String resourcePath, ServletContext servletContext)
    throws ResourceException
  {
    if (resourcePath == null || resourcePath.length() == 0)
    {
      throw new ResourceException("Resource path is empty in "
                                        + getClass().getName()
                                        + ".getResourceLastModified()");
    }

    // check that the resource exists
    //
    URL resourceURL = null;
    try
    {
      resourceURL = servletContext.getResource(RESOURCE_FOLDER + resourcePath);
    }
    catch (MalformedURLException e)
    {
      throw new ResourceException(e);
    }

    if (resourceURL == null)
    {
      throw new ResourceException("Resource [" + resourcePath + "] not found in "
                                        + getClass().getName()
                                        + ".getResourceLastModified()");
    }

    long lastModified = 0;

    try
    {
      URLConnection urlConn = resourceURL.openConnection();
      lastModified = urlConn.getLastModified();
      System.out.println(getClass().getName() + ".getResourceLastModified returning: " + lastModified);
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }

    return lastModified;
  }
}
