package com.jeffhanson.kernel.osgi;

import org.apache.felix.framework.cache.BundleCache;
import org.apache.felix.framework.Felix;
import org.apache.felix.framework.util.FelixConstants;
import org.apache.felix.framework.util.StringMap;
import org.osgi.framework.*;

import java.util.*;
import java.io.File;
import java.io.IOException;

import com.jeffhanson.kernel.*;
import com.jeffhanson.service.SimpleService;

/**
 * User: jhanson
 * Date: Jul 13, 2008
 * Time: 11:55:33 AM
 */
public class OSGiKernel
  implements ServicePollerListener,
             Kernel
{
  private static final int POLL_MILLIS = 30000;

  private KernelActivator m_activator = null;
  private Felix m_felix = null;
  private File m_cachedir = null;
  private ServicePoller m_servicePoller = null;
  private File m_servicesLocation = null;

  /**
   * Constructs an instance of this kernel
   */
  public OSGiKernel()
  {
  }

  public void initialize()
    throws KernelException
  {
    // Create a temporary bundle cache directory
    try
    {
      m_cachedir = File.createTempFile("osgikernel.cache", null);
    }
    catch (IOException e)
    {
      throw new KernelException("Unable to create cache directory: " + e);
    }
    m_cachedir.delete();

    // Create a case-insensitive configuration property map.
    //
    Map configMap = new StringMap(false);

    // Configure the Felix instance to be embedded.
    //
    configMap.put(FelixConstants.EMBEDDED_EXECUTION_PROP, "true");

    // Add core OSGi packages to be exported from the class path
    // via the system bundle.
    //
    configMap.put(Constants.FRAMEWORK_SYSTEMPACKAGES,
                  "org.osgi.framework; version=1.3.0," +
                  "org.osgi.service.packageadmin; version=1.2.0," +
                  "org.osgi.service.startlevel; version=1.0.0," +
                  "org.osgi.service.url; version=1.0.0," +
                  "org.osgi.util.tracker; version=1.3.2," +
                  "com.jeffhanson.service; version=1.0.0");

    // Explicitly specify the directory to use for caching bundles.
    //
    String cacheDir = m_cachedir.getAbsolutePath();
    configMap.put(BundleCache.CACHE_PROFILE_DIR_PROP, cacheDir);
    configMap.put(BundleCache.CACHE_PROFILE_PROP, "OSGiMashupKernel");

    try
    {
      // Create kernel activator;
      //
      m_activator = new KernelActivator();
      List list = new ArrayList();
      list.add(m_activator);

      // Now create an instance of the framework with
      // our configuration properties and activator.
      //
      m_felix = new Felix(configMap, list);

      m_servicePoller = new OSGiServicePoller(m_servicesLocation, POLL_MILLIS);
      m_servicePoller.addServicePollerListener(this);
    }
    catch (Exception e)
    {
      throw new KernelException("Could not create OSGi kernel: " + e);
    }
  }

  /**
   * Utility method used to delete the profile directory when run as
   * a stand-alone application.
   *
   * @param file The file to recursively delete.
   */
  private static void deleteFileOrDir(File file)
  {
    if (file.isDirectory())
    {
      File[] childs = file.listFiles();
      for (int i = 0; i < childs.length; i++)
      {
        deleteFileOrDir(childs[i]);
      }
    }
    file.delete();
  }

  public void setServicesLocation(File serviceLocation)
  {
    m_servicesLocation = serviceLocation;
  }

  public File getServicesLocation()
  {
    return m_servicesLocation;
  }

  /**
   * Retrieves a list of all installed services for this kernel
   *
   * @return a list of all installed services for this kernel
   */
  public String[] getInstalledServiceNames()
  {
    // Use the system bundle activator to gain external
    // access to the set of installed bundles.
    Bundle[] bundles = m_activator.getBundles();
    if (null != bundles && bundles.length > 0)
    {
      ArrayList<String> nameList = new ArrayList<String>();
      for (int i = 0; i < bundles.length; i++)
      {
        Bundle bundle = bundles[i];
        nameList.add(bundle.getSymbolicName());
      }

      String[] names = new String[nameList.size()];
      nameList.toArray(names);
      return names;
    }

    return null;
  }

  /**
   * Retrieves a list of all installed services for this kernel
   *
   * @return a list of all installed services for this kernel
   */
  public Object[] getInstalledServices()
  {
    // Use the system bundle activator to gain external
    // access to the set of installed bundles.
    return m_activator.getBundles();
  }

  /**
   * Retrieves a service object by name
   *
   * @param serviceName - the name of the service to find
   * @return the service object or null
   */
  public Object getServiceByName(String serviceName)
  {
    Bundle[] bundles = m_activator.getBundles();
    if (null != bundles)
    {
      for (int i = 0; i < bundles.length; i++)
      {
        Bundle bundle = bundles[i];
        if (bundle.getSymbolicName().equalsIgnoreCase(serviceName))
        {
          return bundle;
        }
      }
    }
    
    return null;
  }

  /**
   * Starts this kernel
   *
   * @throws KernelException
   */
  public void start()
    throws KernelException
  {
    // Start the felix framework when starting the kernel.
    try
    {
      m_felix.start();
      m_servicePoller.start();
    }
    catch (BundleException e)
    {
      throw new KernelException("Could not start OSGi kernel: " + e);
    }
  }

  /**
   * Stops this kernel
   * 
   * @throws KernelException
   */
  public void stop()
    throws KernelException
  {
    // Stop the felix framework when stopping the kernel.
    try
    {
      m_servicePoller.stop();
      m_felix.stop();
      deleteFileOrDir(m_cachedir);
    }
    catch (BundleException e)
    {
      throw new KernelException("Could not stop OSGi kernel: " + e);
    }
  }

  /**
   * Installs a service in this kernel
   *
   * @param serviceLocation
   * @return Object - the newly install service
   * @throws KernelException
   */
  public Object installService(String serviceLocation)
    throws KernelException
  {
    String tmpServiceLocation = serviceLocation;
    if (tmpServiceLocation.startsWith("file:/") == false)
    {
      tmpServiceLocation = "file:/" + tmpServiceLocation;
    }
    Object retVal = m_activator.installBundle(tmpServiceLocation);
    System.out.println(getClass().getName() + " service installed: " + serviceLocation);
    return retVal;
  }

  /**
   * Uninstalls a service from this kernel
   *
   * @param serviceLocation
   * @throws KernelException
   */
  public void uninstallService(String serviceLocation)
    throws KernelException
  {
    String tmpServiceLocation = serviceLocation;
    if (tmpServiceLocation.startsWith("file:/") == false)
    {
      tmpServiceLocation = "file:/" + tmpServiceLocation;
    }
    m_activator.uninstallBundle(tmpServiceLocation);
    System.out.println(getClass().getName() + " service uninstalled: " + serviceLocation);
  }

  /**
   * Invokes a service installed in this kernel
   *
   * @param service - the service to invoke
   * @param props - any properties needed by the invocation
   * @return the return value from the invocation, or null
   * @throws KernelException
   */
  public Object invokeService(Object service, Properties props)
    throws KernelException
  {
    org.osgi.framework.Bundle bundle = (org.osgi.framework.Bundle)service;
    ServiceReference[] registeredServices = bundle.getRegisteredServices();
    if (null == registeredServices || registeredServices.length <= 0)
    {
      throw new KernelException("Service not registered with bundle");
    }

    BundleContext bundleContext = bundle.getBundleContext();
    if (null == bundleContext)
    {
      throw new KernelException("Bundle context is null");
    }

    Object serviceObj = bundleContext.getService(registeredServices[0]);
    if (null != serviceObj)
    {
      if (serviceObj instanceof SimpleService)
      {
        SimpleService simpleServiceObj = (SimpleService)serviceObj;
        return simpleServiceObj.execute(props);
      }
      else
      {
        throw new KernelException("Service object is invalid " + bundle.getSymbolicName());
      }
    }
    else
    {
      throw new KernelException("Service object is null for " + bundle.getSymbolicName());
    }
  }

  /**
   * ServicePoller method
   *
   * @param evt
   */
  public void serviceAdded(ServicePollerEvent evt)
  {
    try
    {
      installService(evt.getServiceLocation());
    }
    catch (KernelException e)
    {
      e.printStackTrace();
    }
  }

  /**
   * ServicePoller method
   *
   * @param evt
   */
  public void serviceChanged(ServicePollerEvent evt)
  {
    try
    {
      String serviceLocation = evt.getServiceLocation();
      uninstallService(serviceLocation);
      installService(serviceLocation);
    }
    catch (KernelException e)
    {
      e.printStackTrace();
    }
  }

  /**
   * ServicePoller method
   *
   * @param evt
   */
  public void serviceRemoved(ServicePollerEvent evt)
  {
    try
    {
      uninstallService(evt.getServiceLocation());
    }
    catch (KernelException e)
    {
      e.printStackTrace();
    }
  }
}
