package com.jeffhanson.kernel.osgi;

import com.jeffhanson.kernel.ServicePoller;
import com.jeffhanson.kernel.KernelException;
import com.jeffhanson.kernel.ServicePollerListener;
import com.jeffhanson.kernel.ServicePollerEvent;

import java.io.File;
import java.util.*;

/**
 * User: jhanson
 * Date: Jul 14, 2008
 * Time: 8:00:00 PM
 */
public class OSGiServicePoller
  implements ServicePoller
{
  private static class CachedFile
  {
    private File m_file = null;
    private long m_lastModified = 0;

    private CachedFile(File aFile)
    {
      m_file = aFile;
      m_lastModified = aFile.lastModified();
    }
  }

  private File m_servicesLocation = null;
  private ArrayList<CachedFile> m_cachedFiles = null;
  private ArrayList<ServicePollerListener> m_listeners = new ArrayList<ServicePollerListener>();
  private Timer m_timer = null;
  private long m_pollMillis = 0;

  public OSGiServicePoller(File servicesLocation, long pollMillis)
  {
    m_servicesLocation = servicesLocation;
    m_pollMillis = pollMillis;
  }
  
  private void poll()
  {
    File[] files = m_servicesLocation.listFiles();
    if (m_cachedFiles == null)
    {
      // add all new
      if (null != files && files.length > 0)
      {
        m_cachedFiles = new ArrayList<CachedFile>();
        m_cachedFiles.addAll(fileArrayToCachedList(files));
        fireServiceAddedForAll();
      }
    }
    else if (files.length <= 0)
    {
      // remove all
      fireServiceRemovedForAll();
    }
    else
    {
      // pick and choose
      List<File> newFileList = Arrays.asList(files);

      // find deleted files
      ArrayList<File> deletedFiles = new ArrayList<File>();
      Iterator<CachedFile> cachedFileIter = m_cachedFiles.iterator();
      while (cachedFileIter.hasNext())
      {
        CachedFile cachedFile = cachedFileIter.next();
        if (null == isFileInList(cachedFile.m_file, newFileList))
        {
          deletedFiles.add(cachedFile.m_file);
        }
      }

      // Remove deleted files from cache
      Iterator<File> deletedFileIter = deletedFiles.iterator();
      while (deletedFileIter.hasNext())
      {
        CachedFile cachedFile = cachedFileIter.next();
        System.out.println(getClass().getName() + " service removed: " + cachedFile.m_file.getAbsolutePath());
        removeFileFromCache(cachedFile.m_file);
      }

      // remove deleted files from new list
      newFileList.removeAll(deletedFiles);

      // find new and changed files
      Iterator<File> newFileIter = newFileList.iterator();
      while (newFileIter.hasNext())
      {
        File newFile = newFileIter.next();
        CachedFile cachedFile = isCachedFileInList(newFile, m_cachedFiles);
        if (null != cachedFile)
        {
          // test modified date
          if (newFile.lastModified() > cachedFile.m_lastModified)
          {
            System.out.println(getClass().getName() + " service changed: " + newFile.getAbsolutePath());
            fireServiceChanged(newFile);
          }
        }
        else
        {
          m_cachedFiles.add(new CachedFile(newFile));
          System.out.println(getClass().getName() + " service added: " + newFile.getAbsolutePath());
          fireServiceAdded(newFile);
        }
      }
    }
  }

  private List<CachedFile> fileArrayToCachedList(File[] files)
  {
    ArrayList<CachedFile> fileList = new ArrayList<CachedFile>();
    for (int i = 0; i < files.length; i++)
    {
      File file = files[i];
      fileList.add(new CachedFile(file));
    }
    return fileList;
  }

  private void removeFileFromCache(File aFile)
  {
    String aFilePath = aFile.getAbsolutePath();
    ArrayList<CachedFile> removedFiles = new ArrayList<CachedFile>();

    Iterator<CachedFile> cachedFileIter = m_cachedFiles.iterator();
    while (cachedFileIter.hasNext())
    {
      CachedFile cachedFile = cachedFileIter.next();
      if (aFilePath.equalsIgnoreCase(cachedFile.m_file.getAbsolutePath()))
      {
        removedFiles.add(cachedFile);
      }
    }

    Iterator<CachedFile> removedFileIter = removedFiles.iterator();
    while (removedFileIter.hasNext())
    {
      CachedFile removedFile = removedFileIter.next();
      m_cachedFiles.remove(removedFile);
      fireServiceRemoved(removedFile.m_file);
    }
  }

  private File isFileInList(File aFile, List<File> fileList)
  {
    String aFilePath = aFile.getAbsolutePath();
    Iterator<File> iter = fileList.iterator();
    while (iter.hasNext())
    {
      File serviceFile = iter.next();
      String filePath = serviceFile.getAbsolutePath();
      if (filePath.equalsIgnoreCase(aFilePath))
      {
        return serviceFile;
      }
    }

    return null;
  }

  private CachedFile isCachedFileInList(File aFile, List<CachedFile> fileList)
  {
    String aFilePath = aFile.getAbsolutePath();
    Iterator<CachedFile> iter = fileList.iterator();
    while (iter.hasNext())
    {
      CachedFile serviceFile = iter.next();
      String filePath = serviceFile.m_file.getAbsolutePath();
      if (filePath.equalsIgnoreCase(aFilePath))
      {
        return serviceFile;
      }
    }

    return null;
  }

  private void fireServiceRemoved(File aFile)
  {
    ServicePollerEvent evt = new ServicePollerEvent(this, aFile.getAbsolutePath());
    Iterator<ServicePollerListener> listenerIter = m_listeners.iterator();
    while (listenerIter.hasNext())
    {
      ServicePollerListener listener = listenerIter.next();
      listener.serviceRemoved(evt);
    }
  }

  private void fireServiceAdded(File aFile)
  {
    ServicePollerEvent evt = new ServicePollerEvent(this, aFile.getAbsolutePath());
    Iterator<ServicePollerListener> listenerIter = m_listeners.iterator();
    while (listenerIter.hasNext())
    {
      ServicePollerListener listener = listenerIter.next();
      listener.serviceAdded(evt);
    }
  }

  private void fireServiceChanged(File aFile)
  {
    ServicePollerEvent evt = new ServicePollerEvent(this, aFile.getAbsolutePath());
    Iterator<ServicePollerListener> listenerIter = m_listeners.iterator();
    while (listenerIter.hasNext())
    {
      ServicePollerListener listener = listenerIter.next();
      listener.serviceChanged(evt);
    }
  }

  private void fireServiceRemovedForAll()
  {
    Iterator<CachedFile> iter = m_cachedFiles.iterator();
    while (iter.hasNext())
    {
      CachedFile serviceFile = iter.next();
      String filePath = serviceFile.m_file.getAbsolutePath();
      System.out.println(getClass().getName() + " service removed: " + filePath);
      ServicePollerEvent evt = new ServicePollerEvent(this, filePath);
      Iterator<ServicePollerListener> listenerIter = m_listeners.iterator();
      while (listenerIter.hasNext())
      {
        ServicePollerListener listener = listenerIter.next();
        listener.serviceRemoved(evt);
      }
    }

    m_cachedFiles.clear();
  }

  private void fireServiceAddedForAll()
  {
    Iterator<CachedFile> iter = m_cachedFiles.iterator();
    while (iter.hasNext())
    {
      CachedFile serviceFile = iter.next();
      String filePath = serviceFile.m_file.getAbsolutePath();
      System.out.println(getClass().getName() + " service added: " + filePath);
      ServicePollerEvent evt = new ServicePollerEvent(this, filePath);
      Iterator<ServicePollerListener> listenerIter = m_listeners.iterator();
      while (listenerIter.hasNext())
      {
        ServicePollerListener listener = listenerIter.next();
        listener.serviceAdded(evt);
      }
    }
  }

  public void start()
    throws KernelException
  {
    if (m_servicesLocation.exists() == false)
    {
      m_servicesLocation.mkdir();
    }
    else if (!m_servicesLocation.isDirectory())
    {
      throw new KernelException(getClass().getName()
                                + ".start() invalid services location: "
                                + m_servicesLocation.getAbsolutePath());
    }

    TimerTask timerTask = new TimerTask()
    {
      public void run()
      {
        poll();
      }
    };
    m_timer = new Timer();
    m_timer.scheduleAtFixedRate(timerTask, 0, m_pollMillis);
  }

  public void stop()
    throws KernelException
  {
    m_timer.cancel();
    m_timer = null;
  }

  public void addServicePollerListener(ServicePollerListener listener)
  {
    m_listeners.add(listener);
  }

  public void removeServicePollerListener(ServicePollerListener listener)
  {
    m_listeners.remove(listener);
  }
}
