package com.jeffhanson.workflow;

import org.w3c.dom.Document;
import org.w3c.dom.NodeList;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

import javax.xml.parsers.DocumentBuilderFactory;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.io.InputStream;

/**
 * User: jhanson
 * Date: Sep 7, 2008
 * Time: 9:53:22 PM
 */
public class SimpleWorkflow
  implements Workflow
{
  private static class TaskData
  {
    Task task = null;
    String onSuccessTransition = null;
    String onErrorTransition = null;

    private TaskData(Task task,
                     String onSuccessTransition,
                     String onErrorTransition)
    {
      this.task = task;
      this.onSuccessTransition = onSuccessTransition;
      this.onErrorTransition = onErrorTransition;
    }
  }

  private LinkedHashMap<String, TaskData> tasks =
    new LinkedHashMap<String, TaskData>();

  public void initialize(String workflowConfigName)
    throws WorkflowInitializationException
  {
    ClassLoader cl =
      Thread.currentThread().getContextClassLoader();
    if (cl == null)
    {
      cl = SimpleWorkflow.class.getClassLoader();
    }
    
    InputStream inStream =
      cl.getResourceAsStream(workflowConfigName);
    if (null == inStream)
    {
      throw new WorkflowInitializationException(
        "Workflow config [" + workflowConfigName
        + "] not found");
    }

    try
    {
      Document doc =
        DocumentBuilderFactory.newInstance().
          newDocumentBuilder().parse(inStream);
      NodeList taskNodes =
        doc.getElementsByTagName("task");
      if (taskNodes.getLength() <= 0)
      {
        throw new WorkflowInitializationException(
          "No task nodes found in workflow file.");
      }

      for (int i = 0; i < taskNodes.getLength(); i++)
      {
        Node taskNode = taskNodes.item(i);
        NamedNodeMap attributes = taskNode.getAttributes();
        String taskName = XMLUtils.loadAttrValue(attributes,
                                                 "name");
        if (null != taskName && taskName.length() > 0)
        {
          String taskClsName =
            XMLUtils.loadAttrValue(attributes,
                                   "class");
          if (null != taskClsName && taskClsName.length() > 0)
          {
            Class cls = Class.forName(taskClsName);
            if (Task.class.isAssignableFrom(cls) == false)
            {
              throw new WorkflowInitializationException(
                "Class defined in workflow config "
                + "file for task [" + taskName
                + "] is an invalid type.");
            }

            String onSuccessTransition = null;
            String onErrorTransition = null;

            Node onSuccessNode =
              XMLUtils.findFirstNamedChild(taskNode,
                                           "onsuccess");
            if (null != onSuccessNode)
            {
              NamedNodeMap onSuccessNodeAttrs =
                onSuccessNode.getAttributes();
              onSuccessTransition =
                XMLUtils.loadAttrValue(onSuccessNodeAttrs,
                                       "transition");
            }

            Node onErrorNode =
              XMLUtils.findFirstNamedChild(taskNode,
                                           "onerror");
            if (null != onErrorNode)
            {
              NamedNodeMap onErrorNodeAttrs =
                onErrorNode.getAttributes();
              onErrorTransition =
                XMLUtils.loadAttrValue(onErrorNodeAttrs,
                                       "transition");
            }

            Task taskObj = (Task)cls.newInstance();
            taskObj.setName(taskName);
            taskObj.initialize();
            addTask(taskName,
                    taskObj,
                    onSuccessTransition,
                    onErrorTransition);
          }
        }
      }
    }
    catch (Exception e)
    {
      throw new WorkflowInitializationException(e);
    }
  }

  public void execute(WorkflowContext workflowContext)
    throws WorkflowExecutionException
  {
    Iterator<TaskData> iter = tasks.values().iterator();
    if (iter.hasNext())
    {
      TaskData taskData = iter.next();
      while (null != taskData)
      {
        Task task = taskData.task;
        try
        {
          task.execute(workflowContext);
          taskData = tasks.get(taskData.onSuccessTransition);
        }
        catch (Exception e)
        {
          workflowContext.addError(task.getName()
                                   + "-ERROR", e);
          taskData = tasks.get(taskData.onErrorTransition);
        }
      }
    }
  }

  public void addTask(String name, Task task,
                      String onSuccessTransition,
                      String onErrorTransition)
  {
    TaskData taskData = new TaskData(task,
                                     onSuccessTransition,
                                     onErrorTransition);
    tasks.put(name, taskData);
  }

  public Task getTask(String name)
  {
    TaskData taskData = tasks.get(name);
    if (null == taskData)
    {
      return null;
    }
    return taskData.task;
  }

  public Iterator getTaskNames()
  {
    return tasks.keySet().iterator();
  }
}
