package org.mashups4j.kernel.osgi;

import org.mashups4j.kernel.KernelException;

import java.io.File;
import java.util.*;

/**
 * User: jhanson
 * Date: Jul 14, 2008
 * Time: 8:00:00 PM
 */
public class OSGiBundlePoller
  implements BundlePoller
{
  private static class CachedFile
  {
    private File m_file = null;
    private long m_lastModified = 0;

    private CachedFile(File aFile)
    {
      m_file = aFile;
      m_lastModified = aFile.lastModified();
    }
  }

  private File m_bundlesLocation = null;
  private ArrayList<CachedFile> m_cachedFiles = null;
  private ArrayList<BundlePollerListener> m_listeners = new ArrayList<BundlePollerListener>();
  private Timer m_timer = null;
  private long m_pollMillis = 0;

  public OSGiBundlePoller(File bundlesLocation, long pollMillis)
  {
    m_bundlesLocation = bundlesLocation;
    m_pollMillis = pollMillis;
  }
  
  public void poll()
  {
    System.out.println("Polling for bundles at: " + m_bundlesLocation);
    File[] files = m_bundlesLocation.listFiles();
    if (m_cachedFiles == null)
    {
      // add all new
      if (null != files && files.length > 0)
      {
        m_cachedFiles = new ArrayList<CachedFile>();
        m_cachedFiles.addAll(fileArrayToCachedList(files));
        fireBundleAddedForAll();
      }
    }
    else if (files.length <= 0)
    {
      // remove all
      fireBundleRemovedForAll();
    }
    else
    {
      // pick and choose
      List<File> newFileList = Arrays.asList(files);

      // find deleted files
      ArrayList<File> deletedFiles = new ArrayList<File>();
      Iterator<CachedFile> cachedFileIter = m_cachedFiles.iterator();
      while (cachedFileIter.hasNext())
      {
        CachedFile cachedFile = cachedFileIter.next();
        if (null == isFileInList(cachedFile.m_file, newFileList))
        {
          deletedFiles.add(cachedFile.m_file);
        }
      }

      // Remove deleted files from cache
      Iterator<File> deletedFileIter = deletedFiles.iterator();
      while (deletedFileIter.hasNext())
      {
        CachedFile cachedFile = cachedFileIter.next();
        System.out.println(getClass().getName() + " bundle removed: " + cachedFile.m_file.getAbsolutePath());
        removeFileFromCache(cachedFile.m_file);
      }

      // remove deleted files from new list
      newFileList.removeAll(deletedFiles);

      // find new and changed files
      Iterator<File> newFileIter = newFileList.iterator();
      while (newFileIter.hasNext())
      {
        File newFile = newFileIter.next();
        CachedFile cachedFile = isCachedFileInList(newFile, m_cachedFiles);
        if (null != cachedFile)
        {
          // test modified date
          if (newFile.lastModified() > cachedFile.m_lastModified)
          {
            System.out.println(getClass().getName() + " bundle changed: " + newFile.getAbsolutePath());
            fireBundleChanged(newFile);
          }
        }
        else
        {
          m_cachedFiles.add(new CachedFile(newFile));
          System.out.println(getClass().getName() + " bundle added: " + newFile.getAbsolutePath());
          fireBundleAdded(newFile);
        }
      }
    }
  }

  private List<CachedFile> fileArrayToCachedList(File[] files)
  {
    ArrayList<CachedFile> fileList = new ArrayList<CachedFile>();
    for (int i = 0; i < files.length; i++)
    {
      File file = files[i];
      fileList.add(new CachedFile(file));
    }
    return fileList;
  }

  private void removeFileFromCache(File aFile)
  {
    String aFilePath = aFile.getAbsolutePath();
    ArrayList<CachedFile> removedFiles = new ArrayList<CachedFile>();

    Iterator<CachedFile> cachedFileIter = m_cachedFiles.iterator();
    while (cachedFileIter.hasNext())
    {
      CachedFile cachedFile = cachedFileIter.next();
      if (aFilePath.equalsIgnoreCase(cachedFile.m_file.getAbsolutePath()))
      {
        removedFiles.add(cachedFile);
      }
    }

    Iterator<CachedFile> removedFileIter = removedFiles.iterator();
    while (removedFileIter.hasNext())
    {
      CachedFile removedFile = removedFileIter.next();
      m_cachedFiles.remove(removedFile);
      fireBundleRemoved(removedFile.m_file);
    }
  }

  private File isFileInList(File aFile, List<File> fileList)
  {
    String aFilePath = aFile.getAbsolutePath();
    Iterator<File> iter = fileList.iterator();
    while (iter.hasNext())
    {
      File bundleFile = iter.next();
      String filePath = bundleFile.getAbsolutePath();
      if (filePath.equalsIgnoreCase(aFilePath))
      {
        return bundleFile;
      }
    }

    return null;
  }

  private CachedFile isCachedFileInList(File aFile, List<CachedFile> fileList)
  {
    String aFilePath = aFile.getAbsolutePath();
    Iterator<CachedFile> iter = fileList.iterator();
    while (iter.hasNext())
    {
      CachedFile bundleFile = iter.next();
      String filePath = bundleFile.m_file.getAbsolutePath();
      if (filePath.equalsIgnoreCase(aFilePath))
      {
        return bundleFile;
      }
    }

    return null;
  }

  private void fireBundleRemoved(File aFile)
  {
    BundlePollerEvent evt = new BundlePollerEvent(this, aFile.getAbsolutePath());
    Iterator<BundlePollerListener> listenerIter = m_listeners.iterator();
    while (listenerIter.hasNext())
    {
      BundlePollerListener listener = listenerIter.next();
      listener.bundleRemoved(evt);
    }
  }

  private void fireBundleAdded(File aFile)
  {
    BundlePollerEvent evt = new BundlePollerEvent(this, aFile.getAbsolutePath());
    Iterator<BundlePollerListener> listenerIter = m_listeners.iterator();
    while (listenerIter.hasNext())
    {
      BundlePollerListener listener = listenerIter.next();
      listener.bundleAdded(evt);
    }
  }

  private void fireBundleChanged(File aFile)
  {
    BundlePollerEvent evt = new BundlePollerEvent(this, aFile.getAbsolutePath());
    Iterator<BundlePollerListener> listenerIter = m_listeners.iterator();
    while (listenerIter.hasNext())
    {
      BundlePollerListener listener = listenerIter.next();
      listener.bundleChanged(evt);
    }
  }

  private void fireBundleRemovedForAll()
  {
    Iterator<CachedFile> iter = m_cachedFiles.iterator();
    while (iter.hasNext())
    {
      CachedFile bundleFile = iter.next();
      String filePath = bundleFile.m_file.getAbsolutePath();
      System.out.println(getClass().getName() + " bundle removed: " + filePath);
      BundlePollerEvent evt = new BundlePollerEvent(this, filePath);
      Iterator<BundlePollerListener> listenerIter = m_listeners.iterator();
      while (listenerIter.hasNext())
      {
        BundlePollerListener listener = listenerIter.next();
        listener.bundleRemoved(evt);
      }
    }

    m_cachedFiles.clear();
  }

  private void fireBundleAddedForAll()
  {
    Iterator<CachedFile> iter = m_cachedFiles.iterator();
    while (iter.hasNext())
    {
      CachedFile bundleFile = iter.next();
      String filePath = bundleFile.m_file.getAbsolutePath();
      System.out.println(getClass().getName() + " bundle added: " + filePath);
      BundlePollerEvent evt = new BundlePollerEvent(this, filePath);
      Iterator<BundlePollerListener> listenerIter = m_listeners.iterator();
      while (listenerIter.hasNext())
      {
        BundlePollerListener listener = listenerIter.next();
        listener.bundleAdded(evt);
      }
    }
  }

  public void start()
    throws KernelException
  {
    if (m_bundlesLocation.exists() == false)
    {
      m_bundlesLocation.mkdir();
    }
    else if (!m_bundlesLocation.isDirectory())
    {
      throw new KernelException(getClass().getName()
                                + ".start() invalid bundles location: "
                                + m_bundlesLocation.getAbsolutePath());
    }

    TimerTask timerTask = new TimerTask()
    {
      public void run()
      {
        poll();
      }
    };
    m_timer = new Timer();
    m_timer.scheduleAtFixedRate(timerTask, 0, m_pollMillis);
  }

  public void stop()
    throws KernelException
  {
    m_timer.cancel();
    m_timer = null;
  }

  public void addBundlePollerListener(BundlePollerListener listener)
  {
    m_listeners.add(listener);
  }

  public void removeBundlePollerListener(BundlePollerListener listener)
  {
    m_listeners.remove(listener);
  }
}
