package org.mashups4j.kernel.osgi;

import org.apache.felix.framework.cache.BundleCache;
import org.apache.felix.framework.Felix;
import org.apache.felix.framework.util.FelixConstants;
import org.apache.felix.framework.util.StringMap;
import org.osgi.framework.*;

import java.util.*;
import java.io.File;
import java.io.IOException;

import org.mashups4j.kernel.*;
import org.mashups4j.services.SimpleService;

/**
 * User: jhanson
 * Date: Jul 13, 2008
 * Time: 11:55:33 AM
 */
public class OSGiKernel
  implements BundlePollerListener,
             Kernel
{
  private static final int POLL_MILLIS = 30000;
  private static final String API_PATH_PROP = "API_PATH";

  private KernelActivator m_activator = null;
  private Felix m_felix = null;
  private File m_cachedir = null;
  private BundlePoller m_bundlePoller = null;
  private File m_bundlesLocation = null;
  private HashMap<String, Object> serviceMap = new HashMap<String, Object>();

  /**
   * Constructs an instance of this kernel
   */
  public OSGiKernel()
  {
  }

  public void initialize()
    throws KernelException
  {
    // Create a temporary bundle cache directory
    try
    {
      m_cachedir = File.createTempFile("osgikernel.cache", null);
    }
    catch (IOException e)
    {
      throw new KernelException("Unable to create cache directory: " + e);
    }
    m_cachedir.delete();

    // Create a case-insensitive configuration property map.
    //
    Map configMap = new StringMap(false);

    // Configure the Felix instance to be embedded.
    //
    configMap.put(FelixConstants.EMBEDDED_EXECUTION_PROP, "true");
    configMap.put(FelixConstants.SERVICE_URLHANDLERS_PROP, "false");

    // Add core OSGi packages to be exported from the class path
    // via the system bundle.
    //
    configMap.put(Constants.FRAMEWORK_SYSTEMPACKAGES,
                  "org.osgi.framework; version=1.3.0," +
                  "org.osgi.service.packageadmin; version=1.2.0," +
                  "org.osgi.service.startlevel; version=1.0.0," +
                  "org.osgi.service.url; version=1.0.0," +
                  "org.osgi.util.tracker; version=1.3.2," +
                  "org.xml.sax; version=1.0.0," +
                  "sun.misc; verion=1.0.0," +
                  "javax.xml.parsers; version=1.0.0," +
                  "org.w3c.dom; version=1.0.0," +
                  "org.mashups4j.services; version=1.0.0");

    // Explicitly specify the directory to use for caching bundles.
    //
    String cacheDir = m_cachedir.getAbsolutePath();
    configMap.put(BundleCache.CACHE_PROFILE_DIR_PROP, cacheDir);
    configMap.put(BundleCache.CACHE_PROFILE_PROP, "OSGiMashupKernel");

    try
    {
      // Create kernel activator;
      //
      m_activator = new KernelActivator();
      List list = new ArrayList();
      list.add(m_activator);

      // Now create an instance of the framework with
      // our configuration properties and activator.
      //
      m_felix = new Felix(configMap, list);

      m_bundlePoller = new OSGiBundlePoller(m_bundlesLocation, POLL_MILLIS);
      m_bundlePoller.addBundlePollerListener(this);
    }
    catch (Exception e)
    {
      throw new KernelException("Could not create OSGi kernel: " + e);
    }
  }

  /**
   * Utility method used to delete the profile directory when run as
   * a stand-alone application.
   *
   * @param file The file to recursively delete.
   */
  private static void deleteFileOrDir(File file)
  {
    if (file.isDirectory())
    {
      File[] childs = file.listFiles();
      for (int i = 0; i < childs.length; i++)
      {
        deleteFileOrDir(childs[i]);
      }
    }
    file.delete();
  }

  public int getState()
  {
    return m_felix.getState();
  }

  public void setBundlesLocation(File bundlesLocation)
  {
    m_bundlesLocation = bundlesLocation;
  }

  public File getBundlesLocation()
  {
    return m_bundlesLocation;
  }

  /**
   *
   */
  private void resolveServiceMappings()
  {
    serviceMap = new HashMap<String, Object>();

    // Use the system bundle activator to gain external
    // access to the set of installed bundles.
    //
    Bundle[] bundles = m_activator.getBundles();
    if (null != bundles && bundles.length > 0)
    {
      for (int i = 0; i < bundles.length; i++)
      {
        Bundle bundle = bundles[i];

        // get the bundle's ServiceReference list containing all services
        // it has registered or null if it has no registered services
        //
        ServiceReference[] serviceRefs = bundle.getRegisteredServices();
        if (null != serviceRefs)
        {
          for (int j = 0; j < serviceRefs.length; j++)
          {
            ServiceReference serviceRef = serviceRefs[j];
            String path = (String)serviceRef.getProperty(API_PATH_PROP);

            if (null != path && path.length() > 0)
            {
              // Retrieve the service object for the service reference.
              //
              Object service = bundle.getBundleContext().getService(serviceRef);

              //System.out.println("*** resolveServiceMappings found service with path [" + path + "]");

              serviceMap.put(path, service);
            }
            
            // unget the service object to decrement the use count
            //
            bundle.getBundleContext().ungetService(serviceRef);
          }
        }
      }
    }
  }

  /**
   * Retrieves a list of the names for all installed services for this kernel
   *
   * @return a list of all installed services for this kernel
   */
  public String[] getInstalledServiceNames()
  {
    resolveServiceMappings();

    ArrayList<String> nameList = new ArrayList<String>();

    Iterator<String> iter = serviceMap.keySet().iterator();
    while (iter.hasNext())
    {
      String path = iter.next();
      Object service = serviceMap.get(path);
      nameList.add(service.getClass().getName());
    }

    String[] names = new String[nameList.size()];
    nameList.toArray(names);
    return names;
  }

  /**
   * Retrieves a list of the paths for all installed services for this kernel
   *
   * @return a list of names of all installed services for this kernel
   */
  public String[] getInstalledServicePaths()
  {
    resolveServiceMappings();

    String[] paths = new String[serviceMap.keySet().size()];
    serviceMap.keySet().toArray(paths);

    return paths;
  }

  /**
   * Retrieves a list of all installed bundles for this kernel
   *
   * @return a list of all installed bundles for this kernel
   */
  public Object[] getInstalledBundles()
  {
    // Use the system bundle activator to gain external
    // access to the set of installed bundles.
    return m_activator.getBundles();
  }

  /**
   * Retrieves a service object by path
   *
   * @param servicePath - the path of the service to find
   * @return the service object or null
   */
  public Object getServiceByPath(String servicePath)
  {
    resolveServiceMappings();

    Iterator<String> iter = serviceMap.keySet().iterator();
    while (iter.hasNext())
    {
      String path = iter.next();
      if (servicePath.equalsIgnoreCase(path))
      {
        return serviceMap.get(path);
      }
    }

    return null;
  }

  /**
   * Retrieves a service object by name
   *
   * @param serviceName - the name of the service to find
   * @return the service object or null
   */
  public Object getServiceByName(String serviceName)
  {
    resolveServiceMappings();

    Iterator<String> iter = serviceMap.keySet().iterator();
    while (iter.hasNext())
    {
      String path = iter.next();
      Object service = serviceMap.get(path);
      if (serviceName.equalsIgnoreCase(service.getClass().getName()))
      {
        return service;
      }
    }

    return null;
  }

  /**
   * Starts this kernel
   *
   * @throws KernelException
   */
  public void start()
    throws KernelException
  {
    // Start the felix framework when starting the kernel.
    try
    {
      m_felix.start();
      m_bundlePoller.start();
    }
    catch (BundleException e)
    {
      throw new KernelException("Could not start OSGi kernel: " + e);
    }
  }

  /**
   * Stops this kernel
   * 
   * @throws KernelException
   */
  public void stop()
    throws KernelException
  {
    // Stop the felix framework when stopping the kernel.
    try
    {
      m_bundlePoller.stop();
      m_felix.stop();
      deleteFileOrDir(m_cachedir);
    }
    catch (BundleException e)
    {
      throw new KernelException("Could not stop OSGi kernel: " + e);
    }
  }

  /**
   * Installs a bundle in this kernel
   *
   * @param bundleLocation
   * @return Object - the newly install bundle
   * @throws KernelException
   */
  public Object installBundle(String bundleLocation)
    throws KernelException
  {
    String tmpBundleLocation = bundleLocation;
    if (tmpBundleLocation.startsWith("file:/") == false)
    {
      tmpBundleLocation = "file:/" + tmpBundleLocation;
    }
    Bundle installedBundle = m_activator.installBundle(tmpBundleLocation);
    
    return installedBundle;
  }

  /**
   * Uninstalls a bundle from this kernel
   *
   * @param bundleLocation
   * @throws KernelException
   */
  public void uninstallBundle(String bundleLocation)
    throws KernelException
  {
    String tmpBundleLocation = bundleLocation;
    if (tmpBundleLocation.startsWith("file:/") == false)
    {
      tmpBundleLocation = "file:/" + tmpBundleLocation;
    }
    m_activator.uninstallBundle(tmpBundleLocation);
  }

  /**
   * Invokes a service installed in this kernel
   *
   * @param service - the service to invoke
   * @param props - any properties needed by the invocation
   * @return the return value from the invocation, or null
   * @throws KernelException
   */
  public Object invokeService(Object service, Properties props)
    throws KernelException
  {
    if (service instanceof SimpleService)
    {
      SimpleService simpleServiceObj = (SimpleService)service;
      Object retObj = simpleServiceObj.execute(props);
      return retObj;
    }
    else
    {
      throw new KernelException("Service object is invalid: " + service.getClass().getName());
    }
/*
    org.osgi.framework.Bundle bundle = (org.osgi.framework.Bundle)service;
    ServiceReference[] registeredServices = bundle.getRegisteredServices();
    if (null == registeredServices || registeredServices.length <= 0)
    {
      throw new KernelException("Service not registered with bundle");
    }

    BundleContext bundleContext = bundle.getBundleContext();
    if (null == bundleContext)
    {
      throw new KernelException("Bundle context is null");
    }

    Object serviceObj = bundleContext.getService(registeredServices[0]);
    if (null != serviceObj)
    {
      if (serviceObj instanceof SimpleService)
      {
        SimpleService simpleServiceObj = (SimpleService)serviceObj;
        Object retObj = simpleServiceObj.execute(props);
        bundleContext.ungetService(registeredServices[0]);
        return retObj;
      }
      else
      {
        bundleContext.ungetService(registeredServices[0]);
        throw new KernelException("Service object is invalid " + bundle.getSymbolicName());
      }
    }
    else
    {
      throw new KernelException("Service object is null for " + bundle.getSymbolicName());
    }
*/    
  }

  /**
   * BundlePoller method
   *
   * @param evt
   */
  public void bundleAdded(BundlePollerEvent evt)
  {
    try
    {
      String bundleLocation = evt.getBundleLocation();
      if (null != installBundle(bundleLocation))
      {
        System.out.println(getClass().getName() + " service installed: " + bundleLocation);
      }
    }
    catch (KernelException e)
    {
      e.printStackTrace();
    }
  }

  /**
   * BundlePoller method
   *
   * @param evt
   */
  public void bundleChanged(BundlePollerEvent evt)
  {
    try
    {
      String bundleLocation = evt.getBundleLocation();
      uninstallBundle(bundleLocation);
      if (null != installBundle(bundleLocation))
      {
        System.out.println(getClass().getName() + " service changed: " + bundleLocation);
      }
    }
    catch (KernelException e)
    {
      e.printStackTrace();
    }
  }

  /**
   * BundlePoller method
   *
   * @param evt
   */
  public void bundleRemoved(BundlePollerEvent evt)
  {
    try
    {
      String bundleLocation = evt.getBundleLocation();
      uninstallBundle(bundleLocation);
      System.out.println(getClass().getName() + " service uninstalled: " + bundleLocation);
    }
    catch (KernelException e)
    {
      e.printStackTrace();
    }
  }
}
