package org.mashups4j.monitoring;

import java.util.*;

/**
 * User: jhanson
 * Date: Aug 3, 2008
 * Time: 7:54:45 PM
 */
public class LocalProtocolAdapter
  implements EventProtocolAdapter,
             Runnable,
             EventSourceListener
{
  // an event source class that reports time lapses
  //
  private static class ElapsedTimeEventSource
    implements ManagementEventSource
  {
    private static final String EVENT_TYPE = "TIME_ELAPSED";

    private ArrayList<EventCollectorModule> collectors = new ArrayList<EventCollectorModule>();
    private EventReportingAgent eventReportingAgent = null;
    private Date lastTimestamp = new Date();
    private Timer timer = null;

    public void setEventReportingAgent(EventReportingAgent eventReportingAgent)
    {
      this.eventReportingAgent = eventReportingAgent;
    }

    public void addEventCollectorModule(EventCollectorModule eventCollectorModule)
    {
      collectors.add(eventCollectorModule);
    }

    public void removeEventCollectorModule(EventCollectorModule eventCollectorModule)
    {
      collectors.remove(eventCollectorModule);
    }

    public void start()
    {
      TimerTask timerTask = new TimerTask()
      {
        public void run()
        {
          Date currentTimestamp = new Date();
          long elapsedMillis = currentTimestamp.getTime() - lastTimestamp.getTime();
          eventReportingAgent.reportEvent(new EventSourceEvent("ElapsedTimeEventSource",
                                                               new Date(),
                                                               EVENT_TYPE,
                                                               "Elapsed milliseconds: " + elapsedMillis));
          lastTimestamp = currentTimestamp;
        }
      };

      timer = new Timer();
      timer.schedule(timerTask, 5000, 5000);

      System.out.println("ElapsedTimeEventSource started");
    }

    public void stop()
    {
      if (null != timer)
      {
        timer.cancel();
        timer = null;
      }

      System.out.println("ElapsedTimeEventSource stopped");
    }
  }

  // ======================================
  // member fields
  // ======================================
  
  private ArrayList<ManagementEventSource> eventSources = new ArrayList<ManagementEventSource>();
  private ArrayList<ProtocolAdapterListener> listeners = new ArrayList<ProtocolAdapterListener>();
  private boolean running = false;
  private Thread workerThread = null;

  public LocalProtocolAdapter()
  {
    ManagementEventSource eventSource = new ElapsedTimeEventSource();
    
    this.eventSources.add(eventSource);

    // create a reporting agent that reports events as local method calls
    //
    eventSource.setEventReportingAgent(new EventReportingAgent()
    {
      public void reportEvent(EventSourceEvent evt)
      {
        eventOccurred(evt);
      }
    });
  }

  public LocalProtocolAdapter(ManagementEventSource[] eventSources)
  {
    this.eventSources.addAll(Arrays.asList(eventSources));
  }

  public void run()
  {
    running = true;

    System.out.println("LocalProtocolAdapter started");

    while (running)
    {
      Thread.yield();
    }

    System.out.println("LocalProtocolAdapter stopped");

    running = false;
  }

  public void initialize()
    throws EventProtocolAdapterException
  {
  }

  public void addEventReceivedListener(ProtocolAdapterListener listener)
  {
    listeners.add(listener);
  }

  public void removeEventReceivedListener(ProtocolAdapterListener listener)
  {
    listeners.remove(listener);
  }

  public void start()
    throws EventProtocolAdapterException
  {
    Iterator<ManagementEventSource> iter = eventSources.iterator();
    while (iter.hasNext())
    {
      ManagementEventSource eventSource = iter.next();
      eventSource.start();
    }

    workerThread = new Thread(this);
    workerThread.start();
  }

  public void stop()
    throws EventProtocolAdapterException
  {
    Iterator<ManagementEventSource> iter = eventSources.iterator();
    while (iter.hasNext())
    {
      ManagementEventSource eventSource = iter.next();
      eventSource.stop();
    }

    running = false;
  }

  public void eventOccurred(EventSourceEvent evt)
  {
    ProtocolAdapterEvent protocolAdapterEvt = new ProtocolAdapterEvent(evt.getSource(),
                                                                       evt.getTimeStamp(),
                                                                       evt.getType(),
                                                                       evt.getContent());

    Iterator<ProtocolAdapterListener> iter = listeners.iterator();
    while (iter.hasNext())
    {
      ProtocolAdapterListener listener = iter.next();
      listener.eventOccurred(protocolAdapterEvt);
    }
  }
}
