package org.mashups4j.resources;

import javax.servlet.ServletContext;
import javax.servlet.http.HttpServletRequest;
import java.net.URL;
import java.net.MalformedURLException;
import java.net.URLConnection;
import java.io.*;
import java.util.HashMap;

/**
 * User: jhanson
 * Date: Jun 11, 2008
 * Time: 4:29:51 PM
 */
public class FileSystemResourceCache
  implements ResourceCache
{
  private static final String RESOURCE_PREFIX = "WEB-INF/classes";

  private static ResourceCache instance = null;

  // =====================================================
  // public static methods
  // =====================================================

  public static ResourceCache getInstance(boolean saveInstances)
  {
    if (null == instance)
    {
      instance = new FileSystemResourceCache(saveInstances);
    }

    return instance;
  }

  // =====================================================
  // member fields
  // =====================================================

  private boolean saveInstances = false;
  private HashMap<String, Resource> resources = new HashMap<String, Resource>();

  // =====================================================
  // constructors
  // =====================================================

  public FileSystemResourceCache(boolean saveInstances)
  {
    this.saveInstances = saveInstances;
  }

  // =====================================================
  // non-public instance methods
  // =====================================================

  protected String filePathFromRequest(ServletContext servletContext,
                                       String resourcePath)
  {
    return servletContext.getRealPath("") + resourcePath;
  }

  protected void writeToFile(HttpServletRequest request,
                             File file,
                             boolean append)
      throws IOException
  {
    String value = request.getParameter("value");
    if (null == value)
    {
      throw new IOException("Parameter 'value' not found.");
    }

    FileOutputStream outStream = new FileOutputStream(file, append);
    outStream.write(value.getBytes(), 0, value.getBytes().length);
    outStream.flush();
    outStream.close();
  }

  // =====================================================
  // public instance methods
  // =====================================================

  /**
   * Creates a resource
   *
   * @param resourcePath
   * @param context
   * @param data
   * @throws ResourceException
   */
  public void createResource(String resourcePath,
                             Object context,
                             Object data)
    throws ResourceException
  {
    String filePath = filePathFromRequest((ServletContext)context, "/" + RESOURCE_PREFIX + resourcePath);
    File file = new File(filePath);

    try
    {
      boolean append = false;
      writeToFile((HttpServletRequest)data, file, append);
    }
    catch (Exception e)
    {
      e.printStackTrace();
      throw new ResourceException(e);
    }
  }

  /**
   * Retrieves a resource
   * 
   * @param resourcePath
   * @param context
   * @param getCurrentCopy
   * @return the resource
   * @throws ResourceException
   */
  public Resource readResource(String resourcePath,
                               Object context,
                               boolean getCurrentCopy)
    throws ResourceException
  {
    if (resourcePath == null || resourcePath.length() == 0)
    {
      throw new ResourceException("Resource path is empty in "
                                  + getClass().getName()
                                  + ".readResource()");
    }

    String filePath = filePathFromRequest((ServletContext)context, resourcePath);

    // look for cached copy, if possible
    //
    if (!getCurrentCopy)
    {
      Resource resource = resources.get(filePath);
      if (null != resource)
      {
        return resource;
      }
    }

    // get the contents of the resource file
    //
    InputStream inStream = ((ServletContext)context).getResourceAsStream(RESOURCE_PREFIX + resourcePath);
    if (null == inStream)
    {
      throw new ResourceException("Resource [" + resourcePath + "] not found in "
                                  + getClass().getName()
                                  + ".readResource()");
    }

    ByteArrayOutputStream outStream = new ByteArrayOutputStream();
    try
    {
      byte[] bytes = new byte[4096];
      int bytesRead = 0;
      while ((bytesRead = inStream.read(bytes)) > 0)
      {
        outStream.write(bytes, 0, bytesRead);
      }
    }
    catch (IOException e)
    {
      throw new ResourceException("Error retrieving resource [" + resourcePath + "] in "
                                  + getClass().getName()
                                  + ".readResource()");
    }

    byte[] resourceBytes = outStream.toByteArray();

    Resource resource = new SimpleResource(resourcePath, resourceBytes);

    // store copy in cache
    //
    if (saveInstances == true)
    {
      resources.put(filePath, resource);
    }

    return resource;
  }

  /**
   * Modifies a resource
   * 
   * @param resourcePath
   * @param context
   * @param data
   * @throws ResourceException
   */
  public void updateResource(String resourcePath,
                             Object context,
                             Object data)
    throws ResourceException
  {
    String filePath = filePathFromRequest((ServletContext)context, "/" + RESOURCE_PREFIX + resourcePath);
    File file = new File(filePath);

    try
    {
      boolean append = (file.exists() ? true : false);
      writeToFile((HttpServletRequest)data, file, append);
    }
    catch (Exception e)
    {
      e.printStackTrace();
      throw new ResourceException(e);
    }
  }

  /**
   * Deletes a resource
   *
   * @param resourcePath
   * @param context
   * @throws ResourceException
   */
  public void deleteResource(String resourcePath,
                             Object context)
    throws ResourceException
  {
    String filePath = filePathFromRequest((ServletContext)context, "/" + RESOURCE_PREFIX + resourcePath);
    File file = new File(filePath);

    if (file.exists())
    {
      if (file.delete() == false)
      {
        throw new ResourceException("Error deleting resource ["
                                    + filePath + "]");
      }
    }
  }

  /**
   * Retrieves the last-modified date for a resource
   *
   * @param resourcePath
   * @param context
   * @return the last modified date of the resource
   * @throws ResourceException
   */
  public long getResourceLastModified(String resourcePath,
                                      Object context)
    throws ResourceException
  {
    if (resourcePath == null || resourcePath.length() == 0)
    {
      throw new ResourceException("Resource path is empty in "
                                  + getClass().getName()
                                  + ".getResourceLastModified()");
    }

    // check that the resource exists
    //
    URL resourceURL = null;
    
    try
    {
      resourceURL = ((ServletContext)context).getResource(resourcePath);
    }
    catch (MalformedURLException e)
    {
      throw new ResourceException(e);
    }

    if (resourceURL == null)
    {
      throw new ResourceException("Resource [" + resourcePath + "] not found in "
                                  + getClass().getName()
                                  + ".getResourceLastModified()");
    }

    long lastModified = 0;

    try
    {
      URLConnection urlConn = resourceURL.openConnection();
      lastModified = urlConn.getLastModified();
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }

    return lastModified;
  }
}
